unit RRPGUrlPicture;

interface

uses
  Windows, SysUtils, Classes, Graphics, RRPGURLGetter, SyncObjs;

type
  TInnerDesejadoPicture = class(TObject)
  private
    FPicture: TPicture;
    FPictureCarregado: Boolean;
    FTamanhoDesejado: TPoint;
    FContadorReferencia: Integer;
    FCenter: Boolean;
    FStretch: Boolean;
    FProportional: Boolean;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TInnerUrlPicture = class(TObject)
  private
    FOnChange: array of TNotifyEvent;
    FURLGetter: TRRPGURLGetter;
    FPicture: TPicture;
    FDownloading: Boolean;
    FTextoStatus: String;
    FUltimoArquivo: String;
    FPictureLoaded: Boolean;
    FProgresso: Double; // Porcentagem
    FCountReferencia: Integer;
    FMaxBytes: Integer;

    FPicturesDesejados: TList; // TList de TInnerDesejadoPicture

    FURL: String;
  protected
    procedure LiberarPicturesDesejados; virtual;  
    procedure LiberarGetterURL; virtual;
    procedure SetURL(const URL: String); virtual;
    procedure NeedRedraw; virtual;
    procedure CarregarDeArquivo(const Arquivo: String); virtual;
    function CalcularDestRect(Base: TRect; const Stretch, Center, Proporcional: Boolean): TRect; virtual;
    function ObterInnerDesejadoPicture(const Largura, Altura: Integer;
                                       const Center, Stretch, Proporcional: Boolean): TInnerDesejadoPicture; virtual;
    procedure PrepararInnerDesejado(IP: TInnerDesejadoPicture); virtual;
    procedure PrepararTodosInnerDesejados; 

    // Eventos do TPicture
    procedure PictureChanged(Sender: TObject); virtual;

    // Eventos URLGetter
    procedure OnURLErro(Sender: TObject); virtual;
    procedure OnURLDados(const Arquivo: String); virtual;
    procedure OnURLInicio(Sender: TObject); virtual;
    procedure OnURLProgresso(const iPosicao, iMaximo: Integer); virtual;
    procedure OnURLFim(Sender: TObject); virtual;
  public
    constructor Create;
    destructor Destroy; override;

    procedure AdicionarReferencia; virtual;
    procedure RemoverReferencia; virtual;

    procedure AddOnChange(Evento: TNotifyEvent); virtual;
    procedure RemoveOnChange(Evento: TNotifyEvent); virtual;

    procedure DesenharEm(Destino: TCanvas; Regiao: TRect;
                         const Stretch, Center, Proporcional: Boolean;
                         const BoaQualidade: Boolean = True); virtual;

    procedure AddTamanhoDesejado(const Largura, Altura: Integer;
                                 const Center, Stretch, Proporcional: Boolean); virtual;
    procedure RemoveTamanhoDesejado(const Largura, Altura: Integer;
                                    const Center, Stretch, Proporcional: Boolean); virtual;

    property URL: String read FURL write SetURL;
    property Picture: TPicture read FPicture;
    property CountReferencia: Integer read FCountReferencia write FCountReferencia;
    property UltimoArquivo: String read FUltimoArquivo;
  end;

  TURLPicture = class(TObject)
  private
    FOnChange: TNotifyEvent;
    FOnPictureLoadedChange: TNotifyEvent;
    FInner: TInnerUrlPicture;
    FURL: String;
    FMaxBytes: Integer;
    
    FTamanhoDesejado: TPoint;
    FCenterDesejado: Boolean;
    FStretchDesejado: Boolean;
    FProportionalDesejado: Boolean;

    FLastGraphicRef: TGraphic; // Usado apenas para tratar o evento FOnPictureLoadedChange;

    function GetPictureLoaded: Boolean;
    function GetPictureLoadProgresso: Double;
    procedure SetMaxBytes(const Value: Integer);
    function GetDownloading: Boolean;
    function GetUltimoArquivo: String;
  protected
    function GetPicture: TPicture; virtual;
    procedure SetInner(const NovoInner: TInnerUrlPicture); virtual;
    procedure SetURL(const Value: String); virtual;
    procedure OnInnerChange(Sender: TObject); virtual;
  public
    destructor Destroy; override;

    procedure DesenharEm(Destino: TCanvas; Regiao: TRect;
                         Stretch, Center, Proporcional: Boolean;
                         const BoaQualidade: Boolean = True); virtual;

    procedure OtimizarDesenhoPara(const Largura, Altura: Integer;
                                  const Center, Stretch, Proporcional: Boolean); virtual;

    property URL: String read FURL write SetURL;
    property Picture: TPicture read GetPicture;
    property PictureLoaded: Boolean read GetPictureLoaded;
    property PictureLoadProgresso: Double read GetPictureLoadProgresso;
    property MaxBytes: Integer read FMaxBytes write SetMaxBytes;
    property Downloading: Boolean read GetDownloading;
    property UltimoArquivo: String read GetUltimoArquivo;

    // Eventos
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnPictureLoadedChange: TNotifyEvent read FOnPictureLoadedChange write FOnPictureLoadedChange;
  end;

function FindInnerUrlPicture(const URL: String): TInnerUrlPicture;

implementation

uses
  Types, UxTheme, RRPGColorScheme;

threadvar
  LInners: TList;

function FindInnerUrlPicture(const URL: String): TInnerUrlPicture;
var
  L: TList;
  i: Integer;
  Inner: TInnerUrlPicture;
begin
  Result := nil;
  L := LInners;

  if L = nil then
    Exit;

  for i := 0 to L.Count - 1 do
  begin
    Inner := L[i];

    if Inner.URL = URL then
    begin
      Result := Inner;  
      Exit;
    end;
  end;
end;

{ TUrlPicture }

procedure TInnerUrlPicture.AddOnChange(Evento: TNotifyEvent);
var
  N: Integer;
begin
  N := Length(FOnChange);
  SetLength(FOnChange, N + 1);
  FOnChange[N] := Evento;
end;

procedure TInnerUrlPicture.AddTamanhoDesejado(const Largura, Altura: Integer;
  const Center, Stretch, Proporcional: Boolean);
var
  IP: TInnerDesejadoPicture;
begin
  IP := Self.ObterInnerDesejadoPicture(Largura, Altura, Center, Stretch, Proporcional);

  if IP = nil then
  begin
    IP := TInnerDesejadoPicture.Create;
    IP.FTamanhoDesejado.X := Largura;
    IP.FTamanhoDesejado.Y := Altura;
    IP.FCenter := Center;
    IP.FStretch := Stretch;
    IP.FProportional := Proporcional;

    if Self.FPictureLoaded then
      Self.PrepararInnerDesejado(IP);

    FPicturesDesejados.Add(IP);
  end;

  Inc(IP.FContadorReferencia);
end;

procedure TInnerUrlPicture.AdicionarReferencia;
begin
  Inc(FCountReferencia);
end;

function TInnerUrlPicture.CalcularDestRect(Base: TRect;
  const Stretch, Center, Proporcional: Boolean): TRect;
var
  w, h, cw, ch: Integer;
  xyaspect: Double;
begin
  w := Picture.Width;
  h := Picture.Height;
  cw := Base.Right - Base.Left;
  ch := Base.Bottom - Base.Top;

  if Stretch or (Proporcional and ((w > cw) or (h > ch))) then
  begin
    if Proporcional and (w > 0) and (h > 0) then
    begin
      xyaspect := w / h;
      if w > h then
      begin
        w := cw;
        h := Trunc(cw / xyaspect);
        if h > ch then  // woops, too big
        begin
          h := ch;
          w := Trunc(ch * xyaspect);
        end;
      end
      else
      begin
        h := ch;
        w := Trunc(ch * xyaspect);
        
        if w > cw then  // woops, too big
        begin
          w := cw;
          h := Trunc(cw / xyaspect);
        end;
      end;
    end
    else
    begin
      w := cw;
      h := ch;
    end;
  end;

  with Result do
  begin
    Left := 0;
    Top := 0;
    Right := w;
    Bottom := h;
  end;

  if Center then
    OffsetRect(Result, (cw - w) div 2, (ch - h) div 2);

  Result.Left := Result.Left + Base.Left;
  Result.Right := Result.Right + Base.Left;

  Result.Top := Result.Top + Base.Top;
  Result.Bottom := Result.Bottom + Base.Top;    
end;

procedure TInnerUrlPicture.CarregarDeArquivo(const Arquivo: String);
var
  P: TPicture;
  B: TBitmap;
begin
  FPictureLoaded := False;

  try
    P := TPicture.Create;

    try
      P.LoadFromFile(Arquivo);
      B := TBitmap.Create;

      try
        B.Width := P.Width;
        B.Height := P.Height;

        B.Canvas.Draw(0, 0, P.Graphic);

        FPicture.Bitmap := B;
      finally
        B.Free;
      end;
    finally
      P.Free;
    end;
                   // pngimage
    FPictureLoaded := True;

    Self.PrepararTodosInnerDesejados();
    Self.NeedRedraw();
  except
    Self.OnURLErro(Self);
  end;
end;

constructor TInnerUrlPicture.Create;
begin
  FPicture := TPicture.Create;
  FPicture.OnChange := Self.PictureChanged;

  FPicturesDesejados := TList.Create;  

  FURLGetter := TRRPGURLGetter.Create;
  FURLGetter.OnInicio := Self.OnURLInicio;
  FURLGetter.OnErro := Self.OnURLErro;
  FURLGetter.OnProgresso := Self.OnURLProgresso;
  FURLGetter.OnDados := Self.OnURLDados;
  FURLGetter.OnFim := Self.OnURLFim;

  if LInners = nil then
    LInners := TList.Create;
    
  LInners.Add(Self);
end;

procedure TInnerUrlPicture.DesenharEm(Destino: TCanvas; Regiao: TRect;
  const Stretch, Center, Proporcional: Boolean;
  const BoaQualidade: Boolean = True);
var
  Size: TSize;
  R, DestRect: TRect;
  S: string;
  DrawPict: TPicture;
  Largura, LarguraProgresso, Altura, BarraX, BarraY, ClientWidth, ClientHeight,
  AlturaTemp: Integer;
  IP: TInnerDesejadoPicture;

  {$IFNDEF VER150}

  procedure DoBufferedPaint(Canvas: TCanvas);
  var
    MemDC: HDC;
    Rect: TRect;
    PaintBuffer: HPAINTBUFFER;

  begin
    Rect := DestRect;
    PaintBuffer := BeginBufferedPaint(Canvas.Handle, Rect, BPBF_TOPDOWNDIB, nil, MemDC);
    try
      Canvas.Handle := MemDC;
      Canvas.StretchDraw(DestRect, Picture.Graphic);
      BufferedPaintMakeOpaque(PaintBuffer, Rect);
    finally
      EndBufferedPaint(PaintBuffer, True);
    end;
  end;
  {$ENDIF}
begin
  ClientWidth := Regiao.Right - Regiao.Left;
  ClientHeight := Regiao.Bottom - Regiao.Top;

  with Destino do
  begin
    if (FPictureLoaded) and (FPicture <> nil) and
       (FPicture.Graphic <> nil) then
    begin
      Largura := Regiao.Right - Regiao.Left;
      Altura := Regiao.Bottom - Regiao.Top;
      IP := Self.ObterInnerDesejadoPicture(Largura, Altura, Center, Stretch, Proporcional);

      if (IP <> nil) and (IP.FPictureCarregado) and (IP.FPicture.Graphic <> nil) then
      begin
        if IP.FPicture.Graphic is TBitmap then
        begin
          BitBlt(Destino.Handle,
                 Regiao.Left + Largura div 2 - IP.FPicture.Width div 2,
                 Regiao.Top + Altura div 2 - IP.FPicture.Height div 2,
                 IP.FPicture.Width,
                 IP.FPicture.Height,
                 TBitmap(IP.FPicture.Graphic).Canvas.Handle,
                 0, 0,
                 SRCCOPY);
        end else
          Destino.Draw(Regiao.Left + Largura div 2 - IP.FPicture.Width div 2,
                       Regiao.Top + Altura div 2 - IP.FPicture.Height div 2,
                       IP.FPicture.Graphic);
      end else
      begin
        DestRect := Self.CalcularDestRect(Regiao, Stretch, Center, Proporcional);
        DrawPict := FPicture;

        if (DrawPict.Graphic is TBitmap) and (BoaQualidade) then
        begin
          SetStretchBltMode(Destino.Handle, STRETCH_HALFTONE);
          SetBrushOrgEx(Destino.Handle, 0, 0, nil);

          StretchBlt(Destino.Handle, DestRect.Left, DestRect.Top,
                     DestRect.Right - DestRect.Left,
                     DestRect.Bottom - DestRect.Top,
                     TBitmap(DrawPict.Graphic).Canvas.Handle,
                     0, 0, DrawPict.Width, DrawPict.Height,
                     SRCCOPY);
        end else
          StretchDraw(DestRect, DrawPict.Graphic);
      end;
    end else
    begin
      if FDownloading and (FProgresso >= 0) then
      begin
        Largura := Round(ClientWidth * 0.75);
        LarguraProgresso := Round(Largura * FProgresso / 100);
        Altura := 20;

        BarraX := ClientWidth div 2 - Largura div 2 + Regiao.Left;
        BarraY := ClientHeight div 2 - Altura div 2 + Regiao.Top;

        Pen.Width := 1;
        Pen.Style := psSolid;
        Pen.Color := EsquemaAtual.CorTextoNormal;
        Brush.Style := bsClear;
        R := Rect(BarraX, BarraY, BarraX + Largura, BarraY + Altura);
        Rectangle(R);

        InflateRect(R, -1, -1);

        Brush.Style := bsSolid;
        Brush.Color := EsquemaAtual.CorBarraProgresso;
        R.Right := R.Left + LarguraProgresso - 1;

        if R.Right < R.Left then
          R.Right := R.Left;

        Pen.Style := psClear;

        FillRect(R);
      end else
      begin
        S := FTextoStatus;

        if S <> '' then
        begin
          S := '(' + S + ')';
          Size := TextExtent(S);
          R.Left := 0;
          R.Top := 0;
          R.Right := ClientWidth;
          R.Bottom := 0;

          Destino.Refresh;

          Brush.Color := EsquemaAtual.CorWindow;
          Brush.Style := bsSolid;
          Pen.Style := psClear;
          Rectangle(Regiao);

          Windows.SetTextColor(Destino.Handle, EsquemaAtual.CorTextoNormal);
          DrawText(Destino.Handle, PChar(S), Length(S), R, DT_CENTER or DT_CALCRECT or DT_WORDBREAK);

          AlturaTemp := R.Bottom;
          R.Top := Regiao.Top + ClientHeight div 2 - AlturaTemp div 2;
          R.Bottom := R.Top + AlturaTemp;
          R.Left := 0;
          R.Right := Regiao.Right;

          DrawText(Destino.Handle, PChar(S), Length(S), R, DT_CENTER or DT_WORDBREAK);
        end;
      end;
    end;
  end;
end;

destructor TInnerUrlPicture.Destroy;
begin
  Self.LiberarPicturesDesejados();

  Self.LiberarGetterURL();

  if LInners <> nil then
  begin
    LInners.Remove(Self);

    if LInners.Count = 0 then
    begin
      LInners.Free;
      LInners := nil;
    end;
  end;

  FURLGetter.Free;
  FPicturesDesejados.Free;
  FPicture.Free;  
  inherited;
end;

procedure TInnerUrlPicture.LiberarGetterURL;
begin
  if FURLGetter <> nil then
    FURLGetter.PararDownload();

  if FDownloading then
  begin
    FDownloading := False;
    Self.NeedRedraw();
  end;
end;

procedure TInnerUrlPicture.LiberarPicturesDesejados;
var
  i: Integer;
  IP: TInnerDesejadoPicture;
begin
  for i := FPicturesDesejados.Count - 1 downto 0 do
  begin
    IP := FPicturesDesejados[i];

    try
      IP.Free;
    except
    end;
  end;

  FPicturesDesejados.Clear;
end;

procedure TInnerUrlPicture.NeedRedraw;
var
  i: Integer;
begin
  for i := 0 to Length(FOnChange) - 1 do
    FOnChange[i](Self);
end;

function TInnerUrlPicture.ObterInnerDesejadoPicture(const Largura,
  Altura: Integer; const Center, Stretch, Proporcional: Boolean): TInnerDesejadoPicture;
var
  i: Integer;
  IP: TInnerDesejadoPicture;
begin
  Result := nil;

  for i := 0 to FPicturesDesejados.Count - 1 do
  begin
    IP := FPicturesDesejados[i];

    if (IP.FTamanhoDesejado.X = Largura) and
       (IP.FTamanhoDesejado.Y = Altura) and
       (IP.FCenter = Center) and
       (IP.FStretch = Stretch) and
       (IP.FProportional = Proporcional) then
    begin
      Result := IP;
      Exit;
    end;
  end;
end;

procedure TInnerUrlPicture.OnURLDados(const Arquivo: String);
begin
  try
    FUltimoArquivo := Arquivo;
    Self.CarregarDeArquivo(Arquivo);
  finally
    FDownloading := False;
  end;
end;

procedure TInnerUrlPicture.OnURLErro(Sender: TObject);
begin
  FDownloading := False;
  FPictureLoaded := False;
  FTextoStatus := 'Erro ao carregar imagem';
  Self.NeedRedraw();
end;

procedure TInnerUrlPicture.OnURLFim(Sender: TObject);
begin
  FDownloading := False;

  if FURLGetter.Abortado then
    FTextoStatus := 'Download Abortado';

  Self.NeedRedraw();
end;

procedure TInnerUrlPicture.OnURLInicio(Sender: TObject);
begin
  FDownloading := True;
  FTextoStatus := 'Carregando';
  FProgresso := 0;
  Self.NeedRedraw();
end;

procedure TInnerUrlPicture.OnURLProgresso(const iPosicao, iMaximo: Integer);
var
  Prog, Posicao, Maximo: Double;
begin
  FDownloading := True;
  Posicao := iPosicao;
  Maximo := iMaximo;

  FTextoStatus := 'Carregando..';

  if Maximo > 0 then
  begin
    Prog := (Posicao / Maximo) * 100;
    FProgresso := Prog;
  end else
    FProgresso := -1;

  Self.NeedRedraw();

  if iMaximo > 0 then
    if FMaxBytes > 0 then
      if iMaximo > FMaxBytes then
        Abort;  // Abortar, pois o arquivo estourou o tamanho mximo de bytes
end;

procedure TInnerUrlPicture.PictureChanged(Sender: TObject);
begin

end;

procedure TInnerUrlPicture.PrepararInnerDesejado(IP: TInnerDesejadoPicture);
var
  Bit: TBitmap;
  Dest: TRect;
begin
  if FPictureLoaded then
  begin
    if not IP.FPictureCarregado then
    begin
      IP.FPicture.Bitmap := TBitmap.Create;
      Bit := IP.FPicture.Bitmap;
      Dest := Self.CalcularDestRect(Rect(0, 0, IP.FTamanhoDesejado.X, IP.FTamanhoDesejado.Y), False, True, True);

      Bit.Width := Dest.Right - Dest.Left;
      Bit.Height := Dest.Bottom - Dest.Top;

      if FPicture.Graphic is TBitmap then
      begin
        SetStretchBltMode(Bit.Canvas.Handle, STRETCH_HALFTONE);
        SetBrushOrgEx(Bit.Canvas.Handle, 0, 0, nil);

        StretchBlt(Bit.Canvas.Handle, 0, 0,
                   Bit.Width, Bit.Height,
                   TBitmap(FPicture.Graphic).Canvas.Handle,
                   0, 0, FPicture.Width, FPicture.Height,
                   SRCCOPY);
      end else
        Bit.Canvas.StretchDraw(Rect(0, 0, Bit.Width, Bit.Height), FPicture.Graphic);

      IP.FPictureCarregado := True;
    end;
  end else
  begin
    if IP.FPictureCarregado then
    begin
      IP.FPicture.Graphic := nil;
      IP.FPictureCarregado := False;
    end;
  end;
end;

procedure TInnerUrlPicture.PrepararTodosInnerDesejados;
var
  i: Integer;
begin
  for i := 0 to FPicturesDesejados.Count - 1 do
    Self.PrepararInnerDesejado(FPicturesDesejados[i]);
end;

procedure TInnerUrlPicture.RemoveOnChange(Evento: TNotifyEvent);
var
  i, j: Integer;
begin
  for i := Length(FOnChange) - 1 downto 0 do
  begin
    if (TMethod(FOnChange[i]).Data = TMethod(Evento).Data) and
       (TMethod(FOnChange[i]).Code = TMethod(Evento).Code) then
    begin
      for j := i to Length(FOnChange) - 2 do
        FOnChange[j] := FOnChange[j + 1];

      SetLength(FOnChange, Length(FOnChange) - 1);
      Exit;
    end;
  end;
end;

procedure TInnerUrlPicture.RemoverReferencia;
begin
  Dec(FCountReferencia);

  if FCountReferencia <= 0 then
    Self.Free;
end;

procedure TInnerUrlPicture.RemoveTamanhoDesejado(const Largura,
  Altura: Integer; const Center, Stretch, Proporcional: Boolean);
var
  IP: TInnerDesejadoPicture;
begin
  IP := Self.ObterInnerDesejadoPicture(Largura, Altura, Center, Stretch, Proporcional);

  if IP <> nil then
  begin
    Dec(IP.FContadorReferencia);

    if IP.FContadorReferencia <= 0 then
    begin
      FPicturesDesejados.Remove(IP);
      IP.Free;
    end;
  end;
end;

procedure TInnerUrlPicture.SetURL(const URL: String);
begin
  if FURL <> URL then
  begin
    FURL := URL;
    Self.LiberarGetterURL();

    FDownloading := False;
    FTextoStatus := '';

    FPictureLoaded := False;
    FPicture.Graphic := nil;
    Self.NeedRedraw();

    if URL <> '' then
    begin
      Self.PrepararTodosInnerDesejados();
      FURLGetter.CarregarURL(URL);
    end;
  end;
end;

procedure TURLPicture.SetInner(const NovoInner: TInnerUrlPicture);
begin
  if FInner = NovoInner then
    Exit;

  if FInner <> nil then
  begin
    if (FTamanhoDesejado.X > 0) or (FTamanhoDesejado.Y > 0) then
      FInner.RemoveTamanhoDesejado(FTamanhoDesejado.X, FTamanhoDesejado.Y,
                                   FCenterDesejado, FStretchDesejado,
                                   FProportionalDesejado);

    FInner.RemoveOnChange(Self.OnInnerChange);
    FInner.RemoverReferencia();
  end;

  FInner := NovoInner;

  if FInner <> nil then
  begin
    FInner.AdicionarReferencia();
    
    if (FTamanhoDesejado.X > 0) or (FTamanhoDesejado.Y > 0) then
      FInner.AddTamanhoDesejado(FTamanhoDesejado.X, FTamanhoDesejado.Y,
                                FCenterDesejado, FStretchDesejado,
                                FProportionalDesejado);

    FInner.AddOnChange(Self.OnInnerChange);
  end;

  Self.OnInnerChange(Self);
end;

procedure TURLPicture.SetURL(const Value: String);
var
  Temp: TInnerUrlPicture;
begin
  if FURL <> Value then
  begin
    FURL := Value;

    if FURL <> '' then
    begin
      Temp := FindInnerUrlPicture(FURL);

      if Temp <> nil then
      begin
        if FMaxBytes = 0 then
        begin
          if Temp.FMaxBytes <> 0 then
            Temp := nil;
        end else
          if Temp.FMaxBytes = 0 then
            Temp := nil
          else
            if Temp.FMaxBytes > FMaxBytes then
              Temp := nil;      
      end;

      if Temp = nil then
      begin
        Temp := TInnerUrlPicture.Create;
        Temp.URL := FURL;
        Temp.FMaxBytes := FMaxBytes;
      end;

      Self.SetInner(Temp);
    end else
      Self.SetInner(nil);
  end;  
end;

procedure TURLPicture.DesenharEm(Destino: TCanvas; Regiao: TRect;
  Stretch, Center, Proporcional: Boolean;
  const BoaQualidade: Boolean = True);
begin
  if FInner <> nil then
  begin
    if (FTamanhoDesejado.X > 0) and (FTamanhoDesejado.Y > 0) then
    begin
      Regiao.Right := Regiao.Left + FTamanhoDesejado.X;
      Regiao.Bottom := Regiao.Top + FTamanhoDesejado.Y;
      Stretch := FStretchDesejado;
      Center := FCenterDesejado;
      Proporcional := FProportionalDesejado;
    end;

    FInner.DesenharEm(Destino, Regiao, Stretch, Center, Proporcional, BoaQualidade);
  end;
end;

procedure TURLPicture.OnInnerChange(Sender: TObject);
var
  GraphicAtual: TGraphic;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);

  GraphicAtual := nil;

  if FInner <> nil then
    if FInner.Picture <> nil then
      GraphicAtual := FInner.Picture.Graphic;

  if GraphicAtual <> FLastGraphicRef then
  begin
    FLastGraphicRef := GraphicAtual;

    if Assigned(FOnPictureLoadedChange) then
      FOnPictureLoadedChange(Self);
  end;
end;

function TURLPicture.GetPictureLoaded: Boolean;
begin
  if FInner <> nil then
    Result := FInner.FPictureLoaded
  else
    Result := False;
end;

function TURLPicture.GetPictureLoadProgresso: Double;
begin
  if FInner <> nil then
    Result := FInner.FProgresso
  else
    Result := 0;
end;

function TURLPicture.GetUltimoArquivo: String;
begin
  if FInner <> nil then
    Result := FInner.UltimoArquivo
  else
    Result := '';
end;

procedure TURLPicture.SetMaxBytes(const Value: Integer);
begin
  FMaxBytes := Value;
end;

function TURLPicture.GetDownloading: Boolean;
begin
  if FInner <> nil then
    Result := FInner.FDownloading
  else
    Result := False;  
end;

{ TURLPicture }

destructor TURLPicture.Destroy;
begin
  FOnChange := nil;
  FOnPictureLoadedChange := nil;
  Self.SetInner(nil);
  inherited;
end;

function TURLPicture.GetPicture: TPicture;
begin
  if FInner <> nil then
    Result := FInner.Picture
  else
    Result := nil;
end;

procedure TURLPicture.OtimizarDesenhoPara(const Largura, Altura: Integer;
  const Center, Stretch, Proporcional: Boolean);
begin
  if (FTamanhoDesejado.X <> Largura) or
     (FTamanhoDesejado.Y <> Altura) or
     (FCenterDesejado <> Center) or
     (FStretchDesejado <> Stretch) or
     (FProportionalDesejado <> Proporcional) then
  begin
    if Largura < 0 then
      raise Exception.Create('Dimenses invlidas');

    if Altura < 0 then
      raise Exception.Create('Dimenses invlidas');

    if Largura > 0 then
      if Altura <= 0 then
        raise Exception.Create('Dimenses invlidas');

    if Largura > 0 then
      if Altura <= 0 then
        raise Exception.Create('Dimenses invlidas');

    if (FTamanhoDesejado.X > 0) or (FTamanhoDesejado.Y > 0) then
      if FInner <> nil then
        FInner.RemoveTamanhoDesejado(FTamanhoDesejado.X, FTamanhoDesejado.Y,
                                     FCenterDesejado, FStretchDesejado, FProportionalDesejado);

    FTamanhoDesejado.X := Largura;
    FTamanhoDesejado.Y := Altura;
    FCenterDesejado := Center;
    FStretchDesejado := Stretch;
    FProportionalDesejado := Proporcional;

    if (FTamanhoDesejado.X > 0) or (FTamanhoDesejado.Y > 0) then
      if FInner <> nil then
        FInner.AddTamanhoDesejado(FTamanhoDesejado.X, FTamanhoDesejado.Y,
                                  FCenterDesejado, FStretchDesejado, FProportionalDesejado);
  end;
end;

{ TInnerDesejadoPicture }

constructor TInnerDesejadoPicture.Create;
begin
  FPicture := TPicture.Create;
end;

destructor TInnerDesejadoPicture.Destroy;
begin
  FreeAndNil(FPicture);
end;

end.

